/*
   Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
   2011, 2012 Her Majesty the Queen in Right of Canada (Communications
   Research Center Canada)

   Copyright (C) 2025
   Matthias P. Braendli, matthias.braendli@mpb.li

    The Configuration parser sets up the ensemble according
     to the configuration given in a boost property tree, which
     is directly derived from a config file.

     The format of the configuration is given in doc/example.mux
   */
/*
   This file is part of ODR-DabMux.

   ODR-DabMux is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   ODR-DabMux is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ODR-DabMux.  If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include "MuxElements.h"
#include <boost/property_tree/ptree.hpp>
#include <memory>

void parse_ptree(
        boost::property_tree::ptree& pt,
        std::shared_ptr<dabEnsemble> ensemble);

void parse_linkage(
        const boost::optional<boost::property_tree::ptree&> pt_linking,
        std::vector<std::shared_ptr<LinkageSet> >& linkage_sets);

void parse_freq_info(
        const boost::optional<boost::property_tree::ptree&> pt_frequency_information,
        std::vector<FrequencyInformation>& frequency_information);

void parse_other_service_linking(
        const boost::optional<boost::property_tree::ptree&> pt_other_services,
        std::vector<ServiceOtherEnsembleInfo>& service_other_ensemble);
